// KoopmanP examples -- Chapter 8 -- unit tests

// Copyright 2024 Philip Koopman
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "chapter8_examples.h"
#include "support.h"

static const uint32_t initialSeed = SEED;  // initial sum

// Test DualXP checksum versions
void TryDualXP(uint8_t dataWord[], uint32_t dwSize)
{ 
  const uint32_t modulus7  = 125;
  const uint32_t modulus15 = 32749;
  
  // Divisible by 2 data word length
  dwSize &= 0xFFFFFFFE;
  uint32_t resultA = (uint32_t) DualX16P(dataWord, dwSize, modulus7);

  // Divisible by 4 data word length
  dwSize &= 0xFFFFFFFC;
  uint32_t resultB = (uint32_t) DualX32P(dataWord, dwSize, modulus15);
  
  printf("Result DualX-16P=\t0x%02X\n", resultA);
  printf("Result DualX-32P=\t0x%02X\n", resultB);
  
#ifdef VALIDATE
  if(initialSeed == 0)
  {
    assert(resultA == 0x55A1);
    assert(resultB == 0x202943CC);
  } else
if(initialSeed == 0xFF)
  { 
    assert(resultA == 0x5AD3);
    assert(resultB == 0x212849C6);
  } else
if(initialSeed == 0x9D)
  { 
    assert(resultA == 0x75E7);
    assert(resultB == 0x20C6477B);
  } else
  { assert(false);
  }
#endif

}

// Test Koopman checksum versions
void TryKoopmanP(uint8_t dataWord[], uint32_t dwSize)
{ 
  const uint32_t modulus7  = 125;
  const uint32_t modulus15 = 32749;
  const uint32_t modulus23 = 0x7FFFEB;
  const uint32_t modulus31 = 0x7FFFFFED;
  
  uint32_t resultA = (uint32_t) Koopman8P(dataWord, dwSize, modulus7);
  uint32_t resultB = (uint32_t) Koopman16P(dataWord, dwSize, modulus15);
  uint32_t resultC = (uint32_t) Koopman24P(dataWord, dwSize, modulus23);
  uint32_t resultD = (uint32_t) Koopman32P(dataWord, dwSize, modulus31);

  printf("Result Koopman8P=\t0x%02X\n", resultA);
  printf("Result Koopman16P=\t0x%02X\n", resultB);
  printf("Result Koopman24P=\t0x%02X\n", resultC);
  printf("Result Koopman32P=\t0x%02X\n", resultD);
  
#ifdef VALIDATE
  if(initialSeed == 0)
  {
    assert(resultA == 0xCD);
    assert(resultB == 0x1591);
    assert(resultC == 0xFE6AED);
    assert(resultD == 0x6A5C0AE9);
  } else
  if(initialSeed == 0xFF)
  { 
    assert(resultA == 0xF9);
    assert(resultB == 0x9C31);
    assert(resultC == 0xc8A7F1);
    assert(resultD == 0x6B62F2D9);
  } else
  if(initialSeed == 0x9D)
  { 
    assert(resultA == 0xAA);
    assert(resultB == 0xAFEC);
    assert(resultC == 0x1C0B5A);
    assert(resultD == 0x6B2A0378);
  } else
  { assert(false);
  }
#endif

}

// ----------------------
// Run generic modulus implementations and return value; fail if mismatch
int main(void) 
{ 
#ifdef VALIDATE
  const uint32_t dwAlloc = 16; // divisible by 4 and >= dwSize
  uint32_t dwSize = 9;         // divisible by 4
#else
  const uint32_t dwAlloc = 256; // divisible by 4 and >= dwSize
  uint32_t dwSize = 256;        // divisible by 4
#endif
  uint8_t dataWord[dwAlloc];

  InitializeData(dataWord, dwAlloc); // initialize padding bytes too
  PrintWord(dataWord, dwSize);

  printf(">>>>> Initial Seed = %d\n", initialSeed);
  TryKoopmanP(dataWord, dwSize);
  TryDualXP(dataWord, dwSize);
    
  printf("<<<<<<<<<<<<<<<<<<<<<<<<<<< PASS >>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
  return EXIT_SUCCESS;
}


