// Hamming Distance CRC Polynomial Evaluation   
// Copyright 2015-2024, Philip Koopman  koopman@cmu.edu
// Creative Commons Attribution-ShareAlike 4.0 International
//       http://creativecommons.org/licenses/by-sa/4.0/
// No warranty express or implied; use at your own risk
// User assumes responsibility for validating suitability for use
// We suggest using the non-optimized version for validation.

// Fast version with lookup table and hash table

// usage:  hdlen  < polylist
//         hdlen poly
//         hdlen startHD stopHD  < polylist
//         hdlen poly startHD stopHD                        

// stdin is hex CRC polynomial up to 64 bits in implicit +1 notation
// if startHD and stopHD are not specified all HD lengths are computed from 3 up
// Example inputs:

//      ./hdlen 0x82608edb
// gives entire HD profile of CRC-32, which is:
//   0x82608edb {4294967263,91607,2974,268,171,91,57,34,21,12,10,10,10}

//      ./hdlen 0x82608edb 5 7
// gives HD=5, 6, and 7 profile of CRC-32 (skips slow HD computations)
//   0x82608edb {?,?,2974,268,171,?,?,?,?,?,?,?,?}
//  the "?" entries avoid confusion about which weights were computed

// Each example shown in output is a minimum-length codeword at the HD 
//     Example: Len=2975 {0,2215,2866} (0x80000000) (Bits=4)
//  means a 2975 bit long data word with first bit (bit zero) set, bits 2215
//     and bit 2866 set gives a computed CRC result of 0x80000000 for a total
//     of four bits set in that codeword. This is an example demonstrating that
//     CRC-32 0x82608edb fails to provide HD=5 at a 2975-bit data word length. 
//     (it gives only HD=4 due to a codeword with 3 faults in data word and one 
//     fault in the check value.  Check values computed with right-shift algorithm.

// Version 2.0.0 December 2023; updated to include hash table speedup
// Version 2.1.0 March 2024; clean-up and minor speed tweaks
//
// Written for g++ 9.4.0 but beware of portability problems 
//  Compile with:   g++ -O2 hdlen.cpp -o hdlen -std=c++11
// Supports up to 64-bit CRCs, but realistically good large CRCs are going
//   to be too slow to compute to get all HD values
// Note we've seen intermittent problems with higher optimization levels generating code
//       that gives incorrect results on some g++ releases, so recommend staying with -O2

#include <iostream>
#include <cstdio>
#include <sstream>
#include <stdlib.h>
#include <stdint.h>
#include <string>
#include <inttypes.h>
#include <assert.h>
#include <cstring>

using std::hex;
using std::dec;
using std::endl;
using std::cout;
using std::cerr;

#ifdef INSTRUMENT
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif

// Define useful data types
typedef uint64_t poly_t;       // CRC polynomial value, implicit +1 notation
typedef uint64_t posn_t;       // position of a bit within a data word; data word length
typedef uint64_t bigcount_t;   // 64-bit integer for larger counts
typedef uint32_t count_t;      // intentionally only 32-bits count
typedef bool     flag_t;       // Success flag value


static const count_t bitsPerByte = sizeof(uint8_t)*8;
static const count_t polySizeInBits = sizeof(poly_t)*8;  // 64 bits in poly type

static const count_t maxSizeHD  = polySizeInBits+8;  // HD table size
static const posn_t unusedPosn = 0xFFFFffffFFFFffff;    // Unused HD array entry

// How many lookup table entries for LFSR table
//    MUST be an even power of two!!!
//    For small CRCs the initialization automatically reduces table size for init speed
// Can change with compilation parameter, e.g.:   -DMAXLEN=0x200000
#ifndef HASHLEN
#define HASHLEN 0x10000    // power of 2 of 0x40000000 or less
#endif

///////////////////// General purpose helper routines ///////////////

// return number of set bits in a polynomial
static inline count_t BitCount(poly_t value)
{
  // helper lookup table for bit counting  
  // array of # bits in a byte; index is byte value to be evaluated
  static uint8_t bitCountArray[256] =
  { // use int for native machine word size (efficiency) 
    0, 1, 1, 2,  1, 2, 2, 3,  1, 2, 2, 3,  2, 3, 3, 4,
    1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5,
    1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5, 
    2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6,
    1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5,
    2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6,
    2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6,
    3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7,
    1, 2, 2, 3,  2, 3, 3, 4,  2, 3, 3, 4,  3, 4, 4, 5,
    2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6,
    2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6,
    3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7,
    2, 3, 3, 4,  3, 4, 4, 5,  3, 4, 4, 5,  4, 5, 5, 6,
    3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7,
    3, 4, 4, 5,  4, 5, 5, 6,  4, 5, 5, 6,  5, 6, 6, 7,
    4, 5, 5, 6,  5, 6, 6, 7,  5, 6, 6, 7,  6, 7, 7, 8 
  }; // should be 256 entries

  // Count bits in each byte using lookup table
  count_t numBits = 0;
  while (value != 0)  // terminate when all bits counted
  { numBits += bitCountArray[(value) & 0xff];
    value = value >> bitsPerByte;  // runs low to high to optimize for CRCs smaller than 64 bits
  }
  return(numBits);
}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Class HDLen: array that holds the maximum data word length for each HD
class HDLen{
 friend std::ostream& operator <<(std::ostream& hout, const HDLen& h);
 
 private:
   poly_t hPoly = 0;
   posn_t hDLengths[maxSizeHD+1] = {}; // initialize to 0; constructor re-initializes

 public:
   void   SetLen(count_t weightNum, posn_t len);  // Remember HD length
   posn_t GetLen(count_t weightNum);              // Get HD length
   HDLen(poly_t poly);                      // Set currently active polynomial
   ~HDLen();                                // Destructor
};

// Constructor; remember poly and init table
HDLen::HDLen(poly_t poly)
{ hPoly = poly;
  // Init HDLengths table to all unused entries
  for (count_t i = 0; i <= maxSizeHD; i++)
  { hDLengths[i] = unusedPosn;
  }
}

// Destructor
HDLen::~HDLen()
{ // nothing to do for destruction
}

// Set a particular weight number to be a certain dataword length
inline void HDLen::SetLen(count_t weightNum, posn_t len)
{ hDLengths[weightNum] = len;
}

// Retrieve dataword length of a weight number.  unusedPosn if not found yet
inline posn_t HDLen::GetLen(count_t weightNum)
{ return(hDLengths[weightNum]);
}

// Print all computed HD lengths, putting in "?" for uncomputed lengths
std::ostream& operator <<(std::ostream& hout, const HDLen& hval)
{ auto current_flags = hout.flags();

  // print HD Lengths found surrounded by '{' '}'  
  char separator = '{';

  // print feedback polynomial
  hout << hex << "0x" << hval.hPoly << dec << " ";

  // print lengths starting at HD=3; print "?" for any uncomputed length
  count_t weightNum = 3;
  while(weightNum <= BitCount(hval.hPoly)+1)
  { if(hval.hDLengths[weightNum] == unusedPosn)
	  {
      hout << separator << "?"; // ensures '{' and then ',' list separators
	  } else
	  {
      hout << separator << hval.hDLengths[weightNum];
  	}
    separator = ',';
    weightNum++;
  } // end while

  // If we ended up with a null set then print leading {
  if(separator == '{') 
  { hout << "{";
  }
  
  // finally ... end the list
  hout << "}";
  hout.flags(current_flags);
  return hout;
}

//  End of class HDLen
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Class UndetectedClass: list of bit positions in the first undetected code word at HD
class UndetectedClass
{ friend std::ostream& operator <<(std::ostream& uout, const UndetectedClass& uval);

  private:
    poly_t uPoly       = 0; // CRC polynomial
    posn_t uLen        = 0; // Data word length being checked for HD violation
    poly_t uCheckValue = 0; // Check value value of undetected codeword
    posn_t posnList[maxSizeHD]; // bits set in first undetected codeword
                   // subtract 1 from these values to get 0-based bit position
  public:
    UndetectedClass(poly_t p);
    ~UndetectedClass();
    
    void SetCheckValue(poly_t checkValue); // Record check value
    void SetLen(posn_t len);           // Record length for bit positioning
    void SetBitPosn(posn_t posn, count_t posnIndex); // Record codeword bit
    void PrintBits(posn_t dataWordLen);             // Print example codeword
};

// Init undetected bit position data structure
UndetectedClass::UndetectedClass(poly_t p)
{ uPoly = p;
  uCheckValue = uPoly;
  uLen = unusedPosn;
  // init all entries on list to unused
  for (count_t posnIndex=0; posnIndex < maxSizeHD; posnIndex++)
  { posnList[posnIndex] = unusedPosn;
  }
}

UndetectedClass::~UndetectedClass()
{ // Nothing to do for destruction
}
    
// Remember the check value associated with the bits for this HD
void inline UndetectedClass::SetCheckValue(poly_t checkValue)
{ uCheckValue = checkValue; 
} 

// Remember the length, which is needed for proper printing
// This function must be called before printing.
void inline UndetectedClass::SetLen(posn_t len)
{ uLen = len; 
} 

// Set a particular bit position in the codeword bit list
void inline UndetectedClass::SetBitPosn(posn_t posn, count_t posnIndex)
{ posnList[posnIndex-1] = posn;
}

// Print out all bits in undetected codeword plus the check value
std::ostream& operator <<(std::ostream& uout, const UndetectedClass& uval)
{ auto current_flags = uout.flags();
  assert(uval.uLen != unusedPosn); // Must call SetLen first

  count_t bitsSet = 0;
  char separator = '{';  // make separated list elements look pretty

  // print all bit positions that have been recorded
  // Note that k-bit CRC highest possible HD is k; a k+1 bit example is possible
  for(count_t bitPosn = polySizeInBits+1;  bitPosn != 0; bitPosn--)
  { 
    // print only active entries; ignore unused entries
    if(uval.posnList[bitPosn] != unusedPosn) 
    { posn_t thisPosn = uval.uLen - uval.posnList[bitPosn] - 1;
      uout << separator << thisPosn;
      separator = ',';
      bitsSet++;  // tally up the number of entries actually found
    }
  }

  if(separator == '{') {uout << "{";}         // print "{" if empty list
  uout << "} (0x" << hex << uval.uCheckValue << dec << ")";   // print } and check value
  
  // # bits codeword = # bits in dataword (already computed) + # bits in check value 
  bitsSet += BitCount(uval.uCheckValue); 
  uout << " (Bits=" << bitsSet << ")";
  uout.flags(current_flags);
  return uout;
}

// End Class UndetectedClass
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Class LFSR: build and manage an LFSR lookup capability with hash table
//   Uses a hash table sized via compiler define: HASHLEN
class LFSR{
  private:         
     typedef uint32_t rollindex_t;  // Index into rollValues
     typedef uint32_t hashindex_t;  // Index into hashTable

     poly_t      * rollValues = NULL; // LFSR value lookup table by # of rolls
     rollindex_t * hashTable  = NULL; // Stores index into rollValues
     
     rollindex_t nextRollValidIndex = 0;    // First invalid roll table position
     rollindex_t rollArraySize = HASHLEN;   // size of RollValues table
     hashindex_t hashArraySize = HASHLEN*2; // Size of hashTable
     hashindex_t hashMask = hashArraySize-1;  // Mask for hash table wraparound
     
     poly_t myPoly = 0;       // Current CRC polynomial
     uint32_t myPoly32 = 0;   // non-zero if myPoly at most 32 bits
     uint32_t dummyA = 0;     // For data alignment in memory words
     
     inline hashindex_t ComputeHash(poly_t polyVal);
     inline hashindex_t NextHashEntry(hashindex_t hashVal); // probe hash table

#ifdef INSTRUMENT
     bigcount_t hashExtended = 0;     // Statistics for hash table misses
     bigcount_t hashProbes   = 0;     // Statistics for hash table misses
#endif
     
  public:
     // Search for match without using hash table
     posn_t RollSearch(poly_t searchVal, poly_t matchVal, posn_t matchLen, posn_t posn);

     inline poly_t RollByOne(poly_t val);  // advance LFSR 1 bit  
     inline uint32_t RollByOne32(uint32_t val); // 32-bit roll
     
     void AddHashEntry(poly_t polyVal, posn_t len);  // add a new length to tables
     posn_t LookupPosn(poly_t matchVal, posn_t posn); // look up by value and position range
     static const posn_t notFound = 0xFFFFffff;      // Unused hash table entry
     
     LFSR(poly_t newPolynomial, poly_t topBitSet);   // Constructor
     ~LFSR();                                        // Destructor
     LFSR(const LFSR&);			                         // Shallow copy
	   LFSR& operator =(const LFSR&);                  // Shallow copy
};

LFSR& LFSR::operator =(const LFSR&)  = default;
LFSR::LFSR(const LFSR&) = default;

// LFSR Constructor.  Initialize lookup and hash tables
LFSR::LFSR(poly_t newPolynomial, poly_t topBitSet)
{
  assert(BitCount(HASHLEN) == 1);  // HASHLEN must be power of 2
  assert(HASHLEN <= 0x40000000);
  myPoly = newPolynomial;

  // If 32-bit or smaller polynomial, set myPoly32
  if(myPoly <= 0xFFFFFFFF) 
  { myPoly32 = (uint32_t) myPoly;
  }

  // If it is a small CRC, reduce table sizes to save on init cost
  if ( (topBitSet < HASHLEN) && (topBitSet < 0x80000000) )
  { // small CRC; reduce table size
    rollArraySize = (uint32_t) (topBitSet*2);
    hashArraySize = rollArraySize * 2;
    assert(hashArraySize >= topBitSet); // check for overflow issues
    hashMask = hashArraySize-1;
  }
  
  // Allocate LFSR lookup table
  // Do not need to initialize rollValues because nextRollValidIndex tracks valid elements
  rollValues = new poly_t[rollArraySize];
  
  // Allocate LFSR hash table
  hashTable = new rollindex_t[hashArraySize];
  // Initialize with notFound values (all 0xFF bytes)
  std::memset(hashTable, 0xFF, sizeof *hashTable * hashArraySize);
}

// Destructor
LFSR::~LFSR()
{ delete rollValues;
  delete hashTable;
#ifdef INSTRUMENT
  if(hashProbes > 0) { cerr << "###HashProbes=" << hashProbes << endl; }
  if(hashExtended > 0) 
  {   std::cerr << "@@@Hash Table Size Limit Encountered: " << hashExtended << 
      " times -- consider increasing HASHLEN=0x" << 
      hex << HASHLEN << dec << endl; }
#endif  
}

// Roll CRC 1 bit using shift and conditional XOR of the polynomial value
inline poly_t LFSR::RollByOne(poly_t value)
{ if(value & 1)
  { value >>= 1;
    value ^= myPoly;
  } else
  { value >>= 1;
  } 
  return(value);
}

// Roll by 1 specialized to maximum 32-bit polynomial for speed on some systems
inline uint32_t LFSR::RollByOne32(uint32_t value)
{ if(value & 1)
  { value >>= 1;
    value ^= myPoly32;
  } else
  { value >>= 1;
  } 
  return(value);
}

// Hash function is simply taking the bottom bits of LFSR value
//   On average this is better than the fancy computations we tried
inline LFSR::hashindex_t LFSR::ComputeHash(poly_t polyVal)
{ hashindex_t index = (count_t)polyVal;
  index &= hashMask ; 
  return(index);
}

// Hash table probe with wrap-around
inline LFSR::hashindex_t LFSR::NextHashEntry(LFSR::hashindex_t hashVal)
{   hashVal += 1;
    hashVal &= hashMask;
#ifdef INSTRUMENT
    hashProbes++;
#endif
    return(hashVal);
}

// Search for an LFSR value by brute force
posn_t LFSR::RollSearch(poly_t searchVal, poly_t matchVal, posn_t matchLen, posn_t posn)
{
  if(myPoly <= 0xFFFFffffUL)    
  { // Special speedup for 32-bit or smaller polynomial
    // Roll the value forward using only 32-bit values
    uint32_t searchVal32 = (uint32_t)searchVal;
    uint32_t matchVal32  = (uint32_t)matchVal;

    // Check all lengths up to but excluding outer loop bit position  
    // Significant speed boost having specialized small-integer code here    
    while(   (matchLen < posn)
          && (searchVal32 != matchVal32)
         )
    {  
      // Roll for next inner loop bit position (past end of lookup table)
      searchVal32 = RollByOne32(searchVal32);
      matchLen++;
    }  
  } else // CRC polynomial >32 bits in size
  {
    while(   (matchLen < posn)
          && (searchVal != matchVal)
         )
    {  
      // Roll for next inner loop bit position (past end of lookup table)
      searchVal = RollByOne(searchVal);
      matchLen++;
    }  
  }  
  
  // If search ran past posn, no match was found
  if(matchLen >= posn) 
  { matchLen = notFound;}

  return(matchLen);  
}

// Lookup the position of a given LFSR value
//   Only return matches at positions less than "posn"
//   Returns the LFSR position with value "matchVal" if there is one
posn_t LFSR::LookupPosn(poly_t matchVal, posn_t posn)
{ 
  // get first hash table entry for matchVal
  hashindex_t hashVal = ComputeHash(matchVal);
  posn_t matchLen = hashTable[hashVal];
 
  // Probe table until match or end of chain
  //    notFound is maximum value, so will always be > posn
  while ( matchLen < posn )
  { 
    // Make sure it is a true hit and not just a hash collision
    if(rollValues[matchLen]  == matchVal)
    { return(matchLen);
    }
    // Probe for next hash table entry
    hashVal = NextHashEntry(hashVal);
    matchLen = hashTable[hashVal];
  }

  // See if we need to extend to do sequential search  
  if(posn < nextRollValidIndex)
  { // Lookup table had all entries up to posn, so done with search
    return(notFound);
  }

#ifdef INSTRUMENT
  // Do extended length search past lookup table length
  hashExtended++;  // instrumentation that search went past hash table
#endif 
  
  // Start search just beyond highest hash table entry; roll from there
  matchLen = nextRollValidIndex-1;  // last good entry in rollValues
  poly_t searchVal = rollValues[matchLen];

  // Do the manual sequential search
  return(RollSearch(searchVal, matchVal, matchLen, posn));
}

// Add an LFSR entry to the hash table and lookup table based on its position
void LFSR::AddHashEntry(poly_t polyVal, posn_t posn)
{ 
  if (posn < nextRollValidIndex) 
  { // skip if already initialized
    return; 
  } 
  if(posn >= rollArraySize) 
  { // Don't store anything past the size of the lookup table 
    return;
  }
  
  // Do not leave gaps in hash table entries
  ASSERT(posn == nextRollValidIndex);  
  
  hashindex_t hashVal = LFSR::ComputeHash(polyVal);

  while (hashTable[hashVal] != notFound)
  {
    hashVal = NextHashEntry(hashVal);
  }

  rollValues[posn] = polyVal;
  hashTable[hashVal] = (rollindex_t)posn;  
  nextRollValidIndex = (rollindex_t)posn+1;
  return;
}

// End Class HashTable
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
///////////////////// Polynomial Class //////////////////////////////////
class CRCpoly{
  private:         
    poly_t  myPoly = 0;   // implicit +1 polynomial representation
    poly_t  myTopBit = 0; // only the top polynomial bit set
    count_t crcSize = 0;  // Active bits in the poly 1..64, excluding +1
    uint32_t unusedPadding32 = 0;        // alignment padding
    
    HDLen * HDArray = NULL;              // HDLen array for this poly
    UndetectedClass * Undetected = NULL; // Undetected bit array for this poly
    LFSR * myLFSR = NULL;                // LFSR hash and lookup tables
    
    flag_t  myPolyDivXp1 = false;          // flag true if divisible by x+1
	  char unusedPaddingBytes[7] = "";     // alignment padding

    posn_t FindHD(count_t hdGoal);       // Base case for iteration
    posn_t FindHD3();                    // Optimized helper for HD=3 case
    posn_t FindHDOther(count_t hdGoal);  // Optimized helper for HD>=4 case

    flag_t FindHDRecurse(                // Dive deep to find HD general case
                poly_t accum, posn_t posn, count_t hdBitsLeft);
  	inline flag_t CheckSearchLeaf(       // Helper last two bit detection
                poly_t newAccumulator, posn_t posn);
                    
    inline flag_t CheckAccum(            // Helper to check if accum violates HD
                poly_t accum, posn_t posn, count_t hdBitsLeft);

  public:
    inline poly_t Poly();                 // Retrieve binary polynomial value
    inline poly_t TopBitSet();            // Retrieve just top bit in poly
    inline poly_t DivXP1();               // Retrieve div x+1 flag
    
    void   ProcessCRC(                    // Iterate across all HDs
                count_t startHD, count_t maxHD);
				   
    ~CRCpoly();                          // Destructor; release resources
    CRCpoly(poly_t polynomial);          // Constructor
    CRCpoly(const CRCpoly&);			       // Shallow copy
	  CRCpoly& operator =(const CRCpoly&); // Shallow copy
};

// Shallow copy is OK for this class
 CRCpoly& CRCpoly::operator =(const CRCpoly&)  = default;
 CRCpoly::CRCpoly(const CRCpoly&) = default;

// Constructor. Cache results of characterization of poly and set up lists
CRCpoly::CRCpoly(poly_t polynomial)
{ 
  myPoly = polynomial;
  
  // HDArray used to record HD lengths
  HDArray = new HDLen(polynomial);
  assert(HDArray != NULL);
  
  // Undetected array used to record bits in example undetected codeword
  Undetected = new UndetectedClass(polynomial);
  assert(Undetected != NULL);
  
  // Figure out size of polynomial 1..64
  // Count number of bits shifted out until last bit is shifted, leaving zero
  crcSize = 0; 
  while(polynomial != 0)
  { this->crcSize += 1;
    polynomial = polynomial >> 1;  // Shift 0 into unsigned hi bit
  } // end while
  
  // Recreate top bit set now that we know the polynomial size
  // 64-bit polynomial has bit #63 set, so subtract 1 from size
  myTopBit = (poly_t)(1UL) << (crcSize-1); 

  // Determine if divisible by x+1 based on parity of bit count
  count_t polyNumBitsSet = BitCount(myPoly); 
  myPolyDivXp1 = ((polyNumBitsSet & 1) != 0); 
      
  myLFSR = new LFSR(myPoly, myTopBit);
  assert(myLFSR != NULL);
}

// Destructor -- release HDArray and Undetected to avoid memory leaks
CRCpoly::~CRCpoly()
{ delete HDArray;
  delete Undetected;
  delete myLFSR;
}

// Retrieve binary polynomial value
inline poly_t CRCpoly::Poly()
{ return(myPoly);
}

// Retrieve just top bit in poly
inline poly_t CRCpoly::TopBitSet()
{ return(myTopBit);
}

// Retriev div x+1 flag
inline poly_t CRCpoly::DivXP1()
{ return(myPolyDivXp1);
}

// Helper to determine if residual number of bits in a computed check value
//  is small enough to violate the HD.
// Return Value: true means found HD violation; false means no violation
inline flag_t CRCpoly::CheckAccum(poly_t accum, 
                    posn_t posn, count_t hdBitsLeft)
{   
  flag_t successFlag = false;
  
  // Only an accum value with the top bit set can be
  //   the shortest code word at any HD
  if((accum & myTopBit) != 0)
  {
    // See if Accum value has fewer bits than required to hit or exceed the HD
    //   Note: the "-1" is because accum implicitly has one additional bit set
    //     This is done to make the calling code cleaner.
    if (BitCount(accum) < hdBitsLeft-1)
    { 
      successFlag = true;
      // too few bits are set in check value, so this codeword fails to provide HD
      Undetected->SetCheckValue(accum);   // Record check value and this bit position
      Undetected->SetBitPosn(posn, hdBitsLeft);          
    }
  }

  return(successFlag);  // if true, have found the length for this HD
}

// Flatten last two levels of recursion into a search for a 
//   single additional data word fault that leaves only 
//   top bit set in the check value
// Note that all fault bits in data word is never the 
//   limiting case for HD change, because a change requires
//   top bit to be set in the check value
// Return Value: true means found HD violation; false means no violation
inline flag_t CRCpoly::CheckSearchLeaf(poly_t accum, posn_t posn)
{
  const count_t hdBitsLeft = 2;  // only called when 2; used for readability
  
  // When we start we have some data word fault bits set
  //    with the lowest position set being at posn
  // Want to find a codeword bit that differs from current accum value
  //    by having only highest bit inverted,
  //    which would leave that one bit in check value
  const poly_t matchValue = accum ^ myTopBit;          
  posn_t nextPosn = myLFSR->LookupPosn(matchValue, posn);

  flag_t successFlag = false;
  if(nextPosn != LFSR::notFound)
  { 
    successFlag = true;
    // Match for top bit found, exactly giving bits needed to violate HD 
    Undetected->SetCheckValue(matchValue ^ accum);
    Undetected->SetBitPosn(nextPosn, hdBitsLeft);          
  }  
  return(successFlag);   
}        

// Special case for HD=3
// Look for one bit in dataword that results in top bit set in check value
posn_t CRCpoly::FindHD3()
{ 
  // Start with one bit set at the closest bit position to check value
  // Reset the Undetected array
  delete Undetected;
  Undetected = new UndetectedClass(myPoly);
  assert(Undetected != NULL);

  poly_t accum = myPoly;   
  count_t hdGoal = 3;

  // Look for a match with top bit set
  //  Will always happen within 2**k bits for k-bit feedback polynomial  
  posn_t posn = myLFSR->RollSearch(accum, myTopBit, 0, (myTopBit*2)-1);
  
  // Guaranteed success as long as we roll for (2**k)-1 bits
  assert(posn != myLFSR->notFound);
  Undetected->SetCheckValue(TopBitSet());
  Undetected->SetBitPosn(posn, hdGoal);
  return(posn);
}

// Recursively find whether this polynomial at this length provides the specified HD 
// Dives into the middle of a computation with accumulator representing
//   the aggregate contribution to final check value value from previous bits
// Length starts from bit that is the last bit before the check value and works
//   toward longer word lengths. The bit at maxPosn is set from calling level.
// MaxPosn must be 1 or greater   
flag_t CRCpoly::FindHDRecurse(
                         poly_t accum, posn_t maxPosn, count_t hdBitsLeft)
{ 
  ASSERT(maxPosn > 0);
  ASSERT(hdBitsLeft > 3);
  posn_t posn = 0;  
  poly_t rollingValue = myPoly; // first roll position is the polynomial value

  // Posn 0 is special case -- no possibility of recursion because no bits left
  //   But need to check this in case FCS has fewer bits than HD
  poly_t newAccum = rollingValue ^ accum;
  flag_t successFlag = CheckAccum(newAccum, posn, hdBitsLeft);
  
  posn++; // Try the next bit at this level of recursion 

  // Try bit positions 1 and higher while have not violated HD
  while ((posn < maxPosn) && (!successFlag))
  { 
    // Add one more bit fault at this position
    rollingValue = myLFSR->RollByOne(rollingValue);
    newAccum = rollingValue ^ accum;

    // See if this check value is worse than HD with just this one bit set
    if(CheckAccum(newAccum, posn, hdBitsLeft))  
    { successFlag = true;
    }
    // Check for bottom of recursion dive, with 4 HD bits left
    //    1 bit is set in this recursion level
    //    2 more bits set by the leaf routine need to not find a match
    //    1 more bit to hit HD  =>  1+2+1 = 4
    else if (hdBitsLeft == 4)  
    { 
      successFlag = CheckSearchLeaf(newAccum, posn);
    }
    else
    { // Recurse to add contribution of the next bit in the dataword
      //  We know already that posn is 1 or higher
      successFlag = FindHDRecurse(newAccum, posn, hdBitsLeft-1);    
    }
    
    // Don't move posn++ to start of loop -- upsets the first loop bounds logic
    posn++; // Try the next bit at this level of recursion 
  }  // end while
  
  // found an HD violation, so record it here
  if (successFlag)
  { // "posn-1" compensates for posn++ at end of loop instead of start
    Undetected->SetBitPosn(posn-1, hdBitsLeft);  
  }
  
  return(successFlag); // Returns false if HD violation
}


// Find HD for HD>4
posn_t CRCpoly::FindHDOther(count_t hdGoal)
{
  delete Undetected;
  Undetected = new UndetectedClass(myPoly);
  assert(Undetected != NULL);
  
  poly_t accum = myPoly;  // Accumulator walking first bit error
  posn_t posn = 0;
  myLFSR->AddHashEntry(accum, posn);  // ensure posn 0 is set up
    
  // Set up the first bit fault that defines the effective length of the data word
  // Loop stops when the HD is found
  //   The while condition is for 1 bit fault in data word + check value >= HD
  while(!CheckAccum(accum, posn, hdGoal)) // 1 fault bit in data word + bits in check value
  {  
    // Advance the first bit further away from the check value by 1 bit
    // Recursion is skipped for posn 0 because no room for additional bit
    posn++; 
    accum = myLFSR->RollByOne(accum); 
    // Expand lookup tables for this new length
    myLFSR->AddHashEntry(accum, posn);

    // Recurse to see if there is a HD violation at this length
    //    The recursion sets additional bit faults in the data word  
    ASSERT(hdGoal >= 4);
    if( hdGoal == 4)
    { if(CheckSearchLeaf(accum, posn))         { break;} }
    else
    { if(FindHDRecurse(accum, posn, hdGoal-1)) { break; } }
  }

  // found the first error that exceeds HD
  assert(posn != myLFSR->notFound);
  Undetected->SetBitPosn(posn, hdGoal);
  return(posn);
}

// Outer loop to find longest dataword length at a particular HD
posn_t CRCpoly::FindHD(count_t hdGoal)
{ 
  posn_t posn = 0;
  
  if(hdGoal == 3)  // HD=3 special case
  { 
    posn = FindHD3();
  } 
  else if(   ((hdGoal & 1) == 0)    // Can only reuse even-numbered HD results
          && DivXP1()               // Reuse result if divisible by x+1
          && (HDArray->GetLen(hdGoal-1) != unusedPosn)  // and result is available
         )
  { // Skip computation because same as next lower HD for div by x+1
    // note that we do not re-init Undetected because we're going to reuse it
    posn = HDArray->GetLen(hdGoal-1);
  }
  else  // generic case to compute the next HD
  { 
    posn = FindHDOther(hdGoal);  
  }

  // Record the final length of what we found
  HDArray->SetLen(hdGoal,posn);     
  Undetected->SetLen(posn+1);  
 
  cout << "# 0x" << hex << Poly() << dec << "  HD=" << hdGoal;
  if(posn == 0)  // HD=2 on one-bit data word
  { cout << "  NONE  ";
  } else  // have found a length that supports this HD
  { cout << "  len=" << posn << "  "; 
  }
  
  // Print the example showing next longer dataword violates HD
  cout << "Example: Len=" << posn+1 << " ";

  // Exit loop when first bit has found what we are looking for
  cout << *Undetected << endl;
  return(posn);
}

// Print selected range of HDs for a polynomial
void CRCpoly::ProcessCRC(count_t startHD, count_t maxHD)
{
   // All CRCs have HD=2 at infinite length, so clamp starting HD at 3
   if(startHD < 3)  { startHD = 3;}
   
   // All CRCs violate HD= # active bits in polynomial+1  at length=1 (position 0)
   //   So clamp max HD at that number +1 to show the "NONE" entry
   // do something reasonable for invalid maxHD input
   if(   (maxHD > BitCount(myPoly)+2)
      || (maxHD < startHD) 
     )
   { maxHD = BitCount(myPoly)+2; 
   }

   // Status to stderr to track progress when running a batch
   std::cerr << "Poly=0x" << hex << myPoly << dec;
   std::cerr << " startHD=" << startHD << " maxHD=" << maxHD << endl;

   // Find data word lengths for requested HD range
   for (count_t currentHD = startHD; currentHD <= maxHD; currentHD++)
   { FindHD(currentHD);    // all other HDs
   } 
   
   // Print summary findings as last item
   cout << *HDArray << endl;
   
}

//////////////////////////// Main //////////////////////////////////////////////

int main(int argc, char **argv)
{
  poly_t polyValue = 0;      // Working CRC polynomial
  count_t startHD = 0;       // Requested start HD range, if any
  count_t maxHD = 0;         // Requested max HD range, if any

  flag_t useStdin = false;   // Are we getting inputs from STDIN?
  std::istringstream cmd;    // Command line argument stream
  flag_t failFlag = false;   // Command line parsing failure if true

  switch (argc)
  { //  CMD:   ./hdlen   <polyfile.txt
    case 1: // no command line inputs; full HD from stdin
       useStdin = true;
    break;

    //  CMD:   ./hdlen 0x82608edb      
    case 2: // poly as only command line input
       useStdin = false;
       cmd.clear(); cmd.str(argv[1]); if(!(cmd >> hex >> polyValue )) {failFlag =true;}
    break;


    //  CMD:   ./hdlen 5 7  <polyfile.txt     
    case 3: // minHD and maxHD as command line inputs
       useStdin = true;
       cmd.clear(); cmd.str(argv[1]); if(!(cmd >> dec >> startHD   )) {failFlag =true;}
       cmd.clear(); cmd.str(argv[2]); if(!(cmd >> dec >> maxHD     )) {failFlag =true;}
    break;

    //  CMD:   ./hdlen 0x82608edb 5 7      
    case 4: // poly, minHD and maxHD as command line inputs
       useStdin = false;
       cmd.clear(); cmd.str(argv[1]); if(!(cmd >> hex >> polyValue )) {failFlag =true;}
       cmd.clear(); cmd.str(argv[2]); if(!(cmd >> dec >> startHD   )) {failFlag =true;}
       cmd.clear(); cmd.str(argv[3]); if(!(cmd >> dec >> maxHD     )) {failFlag =true;}
    break;

    default: // invalid number of command line parameters
       failFlag = true;
    break;      
  } // end of switch

  if(failFlag)
  { std::cerr << "Usage: " << argv[0]  
            << " [poly] | [StartHD MaxHD] | [poly StartHD MaxHD]" << endl;
  } else
  { do //  do ... while to ensure always process at least one poly if available 
    { 
      // Get next polynomial to analyze
      if(useStdin)
      { std::cin >> hex >> polyValue >> dec;  
        if (std::cin.fail()) { break; }
      }

      // Perform complete computation for one polynomial
      CRCpoly * Poly = new CRCpoly(polyValue);
      assert(Poly != NULL);

      Poly->ProcessCRC(startHD, maxHD);  
      delete Poly;
      
    } while ((!feof(stdin)) && useStdin);   // end while
  }

  return(failFlag);
}